from pathlib import Path
from threading import Thread
from time import sleep
from tkinter import filedialog

import customtkinter
from rich import print

from metaDMG.cli import cli_utils


#%%

# Modes: "System" (standard), "Dark", "Light"
customtkinter.set_appearance_mode("dark")
# Themes: "blue" (standard), "green", "dark-blue"
customtkinter.set_default_color_theme("blue")

FONT = customtkinter.ThemeManager.theme["text"]["font"]


#%%


def format_directory(path: Path):
    return path.name + "/"


def path_to_text(filepath: str, cut: int = 10) -> str:
    text = Path(filepath).name
    if len(text) > cut:
        text = text[:cut] + " ..."
    return text


#%%

KW_OPTIONS_MENU = dict(
    fg_color="gray30",
    button_color="gray50",
    button_hover_color="gray60",
)

KW_OPTIONS_MENU_DISABLED = dict(
    fg_color="gray25",
    button_color="gray30",
)


KW_BUTTON_BASE = dict(
    hover=True,
    border_width=2,
    text_color_disabled="gray40",
)

KW_BUTTON_GOOD_COLORS = dict(
    fg_color=None,
    hover_color="gray30",
    border_color="gray30",
)

KW_BUTTON_BAD_COLORS = dict(
    hover_color="#78413d",
    border_color="#78413d",
    fg_color="#603431",
)


KW_ENTRY_GOOD_COLORS = dict(
    border_color="gray30",
    text_color="gray90",
    placeholder_text_color="gray90",
    fg_color="#2A2D2E",
)
KW_ENTRY_DISABLED = dict(
    border_color="gray30",
    text_color="gray40",
    placeholder_text_color="gray40",
    fg_color="#2A2D2E",
)

KW_ENTRY_BAD_COLORS = dict(
    border_color="#78413d",
    text_color="gray90",
    placeholder_text_color="gray90",
    fg_color="#603431",
)


KW_SWITCH = dict(button_color="gray80")
KW_SWITCH_DISABLED = dict(button_color="gray40")


KW_SLIDER = dict(
    button_color="#1F6AA5",
    progress_color="#AAB0B5",
)
KW_SLIDER_DISABLED = dict(
    button_color="gray40",
    progress_color="gray40",
)


TOP_LABEL_WIDTH = 50
TOP_LABEL_PAD_X = 20
TOP_LABEL_GRID_KW = dict(
    pady=12,
    padx=TOP_LABEL_PAD_X,
    sticky="nsw",
)

CENTER_LABEL_WIDTH = 50
CENTER_LABEL_PAD_X = 20
CENTER_LABEL_GRID_KW = dict(
    pady=12,
    padx=CENTER_LABEL_PAD_X,
    sticky="nsw",
)

RIGHT_LABEL_WIDTH = 50
RIGHT_LABEL_PAD_X = 20
RIGHT_LABEL_GRID_KW = dict(
    pady=12,
    padx=RIGHT_LABEL_PAD_X,
    sticky="nsw",
)


#%%


class Gui(customtkinter.CTk):
    def __init__(self, verbose: bool = True):
        super().__init__()

        self.verbose = verbose

        self.title("MetaDMG Configuirator")

        WIDTH = 785
        HEIGHT = 655
        X, Y = self.get_center_coordinates(WIDTH, HEIGHT)
        self.geometry(f"{WIDTH}x{HEIGHT}+{X}+{Y}")
        self.resizable(False, False)

        # call .on_closing() when app gets closed
        self.protocol("WM_DELETE_WINDOW", self.on_closing)

        # ============ create frames ============
        self.setup_header()

        # ============ frame_top ============
        self.setup_top_frame()

        # ============ frame_center ============
        self.setup_center_frame()

        # ============ frame_right ============
        self.setup_right_frame()

        # ============ background threads ============
        self.start_similarity_scores_background_thread()
        self.start_minimum_reads_background_thread()

    # ============ HEADER ============

    def setup_header(self):

        self.headline = customtkinter.CTkLabel(
            master=self,
            justify="center",
            text="Confi-gui-rator",
            text_font=(FONT, "-30"),
        )
        self.headline.grid(
            row=0,
            column=0,
            columnspan=2,
        )

    # ============ TOP FRAME ============

    def setup_top_frame(self):

        self.frame_top = customtkinter.CTkFrame(
            master=self,
        )
        self.frame_top.grid(
            row=1,
            column=0,
            sticky="nswe",
            padx=20,
            pady=20,
        )

        # # empty row with minsize as spacing
        # self.frame_top.grid_rowconfigure(0, minsize=50)
        # # empty row with minsize as spacing
        self.frame_top.grid_columnconfigure(1, minsize=80)
        self.TOP_LABEL_KW = dict(master=self.frame_top, width=TOP_LABEL_WIDTH)

        self.setup_bam_file()
        self.setup_damage_mode()

    # ============ BAM FILE ============

    def setup_bam_file(self):

        self.bam_label = customtkinter.CTkLabel(
            text="BAM file:",
            **self.TOP_LABEL_KW,
        )
        self.bam_label.grid(
            row=0,
            column=0,
            **TOP_LABEL_GRID_KW,
        )

        self.bam_file_path = None

        self.bam_button = customtkinter.CTkButton(
            master=self.frame_top,
            text="Select file",
            command=self.bam_callback,
            **KW_BUTTON_BASE,
            **KW_BUTTON_BAD_COLORS,
        )
        self.bam_button.grid(
            row=0,
            column=2,
            pady=12,
            sticky="e",
        )
        self.bam_is_okay = False

    # ============ DAMAGE MODE ============

    def setup_damage_mode(self):

        self.damage_mode_label = customtkinter.CTkLabel(
            text="Damage Mode:",
            **self.TOP_LABEL_KW,
        )
        self.damage_mode_label.grid(
            row=1,
            column=0,
            **TOP_LABEL_GRID_KW,
        )

        self.damage_mode_value = cli_utils.DAMAGE_MODE.LCA

        self.damage_mode_menu = customtkinter.CTkOptionMenu(
            master=self.frame_top,
            values=cli_utils.DAMAGE_MODE.list(),
            command=self.damage_mode_collback,
            **KW_OPTIONS_MENU,
        )

        self.damage_mode_menu.grid(
            row=1,
            column=2,
            pady=12,
            sticky="e",
        )

        self.is_lca = True

    # ============ CENTER FRAME ============

    def setup_center_frame(self):

        self.frame_center = customtkinter.CTkFrame(
            master=self,
        )
        self.frame_center.grid(
            row=2,
            column=0,
            sticky="nswe",
            padx=20,
            # pady=20,
        )

        # configure grid layout (3 x 2) (row x column)
        self.frame_center.grid_rowconfigure(0, minsize=10)
        self.CENTER_LABEL_KW = dict(master=self.frame_center, width=CENTER_LABEL_WIDTH)

        self.setup_names()
        self.setup_nodes()
        self.setup_acc2tax()
        self.setup_similarity_score()
        self.setup_min_mapping_quality()
        self.setup_custom_database()
        self.setup_lca_rank()

    # ============ NAMES ============

    def setup_names(self):

        # self.names_file_string = customtkinter.StringVar()
        self.names_file_path = None

        self.names_label = customtkinter.CTkLabel(
            text="Names:",
            **self.CENTER_LABEL_KW,
        )
        self.names_label.grid(
            row=1,
            column=0,
            **CENTER_LABEL_GRID_KW,
        )

        self.names_button = customtkinter.CTkButton(
            master=self.frame_center,
            text="Select file",
            command=self.names_callback,
            **KW_BUTTON_BASE,
            **KW_BUTTON_BAD_COLORS,
        )
        self.names_button.grid(
            row=1,
            column=1,
            pady=12,
            padx=10,
        )
        self.names_is_okay = False
        self.names_button_colors = KW_BUTTON_BAD_COLORS

    # ============ NODES ============

    def setup_nodes(self):

        self.nodes_file_path = None

        self.nodes_label = customtkinter.CTkLabel(
            text="Nodes:",
            **self.CENTER_LABEL_KW,
        )
        self.nodes_label.grid(
            row=2,
            column=0,
            **CENTER_LABEL_GRID_KW,
        )

        self.nodes_button = customtkinter.CTkButton(
            master=self.frame_center,
            text="Select file",
            command=self.nodes_callback,
            **KW_BUTTON_BASE,
            **KW_BUTTON_BAD_COLORS,
        )
        self.nodes_button.grid(
            row=2,
            column=1,
            pady=12,
            padx=10,
        )

        self.nodes_is_okay = False
        self.nodes_button_colors = KW_BUTTON_BAD_COLORS

    # ============ ACC2TAX ============

    def setup_acc2tax(self):

        self.acc2tax_file_path = None

        self.acc2tax_label = customtkinter.CTkLabel(
            text="Acc2tax:",
            **self.CENTER_LABEL_KW,
        )
        self.acc2tax_label.grid(
            row=3,
            column=0,
            **CENTER_LABEL_GRID_KW,
        )

        self.acc2tax_button = customtkinter.CTkButton(
            master=self.frame_center,
            text="Select file",
            command=self.acc2tax_callback,
            **KW_BUTTON_BASE,
            **KW_BUTTON_BAD_COLORS,
        )

        self.acc2tax_button.grid(
            row=3,
            column=1,
            pady=12,
            padx=10,
        )

        self.acc2tax_is_okay = False
        self.acc2tax_button_colors = KW_BUTTON_BAD_COLORS

    # ============ SIMILARITY SCORE ============

    def setup_similarity_score(self):

        self.similarity_score_label = customtkinter.CTkLabel(
            text="Similarity Score:",
            **self.CENTER_LABEL_KW,
        )
        self.similarity_score_label.grid(
            row=4,
            column=0,
            **CENTER_LABEL_GRID_KW,
        )

        self.frame_similarity = customtkinter.CTkFrame(
            master=self.frame_center,
            fg_color="#2A2D2E",
        )
        self.frame_similarity.grid(
            row=4,
            column=1,
            columnspan=3,
            rowspan=1,
            # pady=20,
            padx=30,
            sticky="w",
        )

        self.similarity_score_min_value = customtkinter.StringVar(
            value="0.95",
        )

        self.similarity_score_min = customtkinter.CTkEntry(
            master=self.frame_similarity,
            textvariable=self.similarity_score_min_value,
            width=40,
            **KW_ENTRY_GOOD_COLORS,
        )
        self.similarity_score_min.grid(
            row=0,
            column=0,
            # sticky="nse",
        )

        self.similarity_score_to = customtkinter.CTkLabel(
            master=self.frame_similarity,
            text="to",
            width=20,
        )
        self.similarity_score_to.grid(
            row=0,
            column=1,
            # sticky="ns",
        )

        self.similarity_score_max_value = customtkinter.StringVar(
            value="1.0",
        )

        self.similarity_score_max = customtkinter.CTkEntry(
            master=self.frame_similarity,
            textvariable=self.similarity_score_max_value,
            width=40,
            **KW_ENTRY_GOOD_COLORS,
        )
        self.similarity_score_max.grid(
            row=0,
            column=2,
            # sticky="nsw",
        )

    # ============ MIN MAPPING QUALITY ============

    def setup_min_mapping_quality(self):

        MIN_MAPPING_QUALITY_MIN = 0
        MIN_MAPPING_QUALITY_DEFAULT = 0
        MIN_MAPPING_QUALITY_MAX = 50

        self.min_mapping_quality_label = customtkinter.CTkLabel(
            text="Minimum Mapping Quality:",
            **self.CENTER_LABEL_KW,
        )
        self.min_mapping_quality_label.grid(
            row=5,
            column=0,
            **CENTER_LABEL_GRID_KW,
        )

        self.frame_min_mapping_quality = customtkinter.CTkFrame(
            master=self.frame_center,
            fg_color="#2A2D2E",
        )
        self.frame_min_mapping_quality.grid(
            row=5,
            column=1,
            sticky="e",
        )

        self.min_mapping_quality_slider = customtkinter.CTkSlider(
            master=self.frame_min_mapping_quality,
            command=self.min_mapping_quality_slider_callback,
            from_=MIN_MAPPING_QUALITY_MIN,
            to=MIN_MAPPING_QUALITY_MAX,
            number_of_steps=MIN_MAPPING_QUALITY_MAX - MIN_MAPPING_QUALITY_MIN,
            width=100,
            **KW_SLIDER,
        )
        self.min_mapping_quality_slider.set(MIN_MAPPING_QUALITY_DEFAULT)
        self.min_mapping_quality_slider.grid(
            row=0,
            column=0,
            sticky="e",
        )

        self.min_mapping_quality_value = customtkinter.CTkLabel(
            master=self.frame_min_mapping_quality,
            width=20,
        )
        self.min_mapping_quality_slider_callback(MIN_MAPPING_QUALITY_DEFAULT)
        self.min_mapping_quality_value.grid(
            row=0,
            column=1,
            pady=12,
            padx=0,
        )

    # ============ CUSTOM DATABASE ============

    def setup_custom_database(self):

        self.custom_database_label = customtkinter.CTkLabel(
            text="Custom Database:",
            **self.CENTER_LABEL_KW,
        )
        self.custom_database_label.grid(
            row=6,
            column=0,
            **CENTER_LABEL_GRID_KW,
        )

        self.custom_database_bool = customtkinter.BooleanVar()
        self.custom_database_button_title = customtkinter.StringVar(
            value="False",
        )
        self.custom_database_switch = customtkinter.CTkSwitch(
            master=self.frame_center,
            textvariable=self.custom_database_button_title,
            command=self.custom_database_callback,
            onvalue=True,
            offvalue=False,
            **KW_SWITCH,
        )
        self.custom_database_switch.deselect()
        self.custom_database_switch.grid(
            row=6,
            column=1,
        )

    # ============ LCA RANK ============

    def setup_lca_rank(self):

        self.lca_rank_label = customtkinter.CTkLabel(
            text="LCA Rank:",
            **self.CENTER_LABEL_KW,
        )
        self.lca_rank_label.grid(
            row=7,
            column=0,
            **CENTER_LABEL_GRID_KW,
        )

        self.lca_rank_string = customtkinter.StringVar(
            value=cli_utils.RANKS.none,
        )

        self.lca_rank_menu = customtkinter.CTkOptionMenu(
            master=self.frame_center,
            values=cli_utils.RANKS.str_list(),
            command=self.lca_rank_callback,
            text_color_disabled="gray40",
            **KW_OPTIONS_MENU,
        )
        self.lca_rank_menu.set("none")  # set initial value

        self.lca_rank_menu.grid(
            row=7,
            column=1,
        )

    # ============ RIGHT FRAME ============

    def setup_right_frame(self):

        self.frame_right = customtkinter.CTkFrame(
            master=self,
        )
        self.frame_right.grid(
            row=1,
            column=1,
            # sticky="e",
            # sticky="nswe",
            rowspan=3,
            # padx=20,
            pady=20,
        )

        # configure grid layout (3 x 2) (row x column)

        self.frame_right.grid_rowconfigure(0, minsize=10)
        self.RIGHT_LABEL_KW = dict(
            master=self.frame_right,
            width=RIGHT_LABEL_WIDTH,
        )
        self.setup_max_position()
        self.setup_min_reads()
        self.setup_bayesian()
        self.setup_forward()
        self.setup_parallel_samples()
        self.setup_cores_per_sample()
        self.setup_prefix()
        self.setup_suffix()
        self.setup_long_name()
        self.setup_config_name()
        self.setup_output_dir()
        self.setup_save_config()

    # ============ MAX POSITION ============

    def setup_max_position(self):

        MAX_POSITION_MIN = 1
        MAX_POSITION_DEFAULT = 15
        MAX_POSITION_MAX = 30

        self.max_position_label = customtkinter.CTkLabel(
            text="Maximum Position:",
            **self.RIGHT_LABEL_KW,
        )
        self.max_position_label.grid(
            row=1,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.frame_max_position = customtkinter.CTkFrame(
            master=self.frame_right,
            fg_color="#2A2D2E",
        )
        self.frame_max_position.grid(
            row=1,
            column=1,
            sticky="e",
        )

        self.max_position_slider = customtkinter.CTkSlider(
            master=self.frame_max_position,
            command=self.max_position_slider_callback,
            from_=MAX_POSITION_MIN,
            to=MAX_POSITION_MAX,
            number_of_steps=MAX_POSITION_MAX - MAX_POSITION_MIN,
            width=100,
            **KW_SLIDER,
        )
        self.max_position_slider.set(MAX_POSITION_DEFAULT)
        self.max_position_slider.grid(
            row=0,
            column=0,
            sticky="e",
        )

        self.max_position_value = customtkinter.CTkLabel(
            master=self.frame_max_position,
            width=20,
        )
        self.max_position_value.grid(
            row=0,
            column=1,
            pady=12,
            padx=0,
        )

        self.max_position_slider_callback(MAX_POSITION_DEFAULT)

    # ============ MINIMUM READS ============

    def setup_min_reads(self):

        self.min_reads_label = customtkinter.CTkLabel(
            text="Minimum number of reads:",
            **self.RIGHT_LABEL_KW,
        )
        self.min_reads_label.grid(
            row=2,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.min_reads_value = customtkinter.StringVar(
            value="0",
        )

        self.min_reads_entry = customtkinter.CTkEntry(
            master=self.frame_right,
            textvariable=self.min_reads_value,
            **KW_ENTRY_GOOD_COLORS,
        )
        self.min_reads_entry.grid(
            row=2,
            column=1,
        )

    # ============ BAYESIAN ============

    def setup_bayesian(self):

        self.bayesian_label = customtkinter.CTkLabel(
            text="Bayesian:",
            **self.RIGHT_LABEL_KW,
        )
        self.bayesian_label.grid(
            row=3,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.bayesian_bool = customtkinter.BooleanVar()
        self.bayesian_button_title = customtkinter.StringVar(
            value="False",
        )
        self.bayesian_switch = customtkinter.CTkSwitch(
            master=self.frame_right,
            textvariable=self.bayesian_button_title,
            command=self.bayesian_callback,
            onvalue=True,
            offvalue=False,
            **KW_SWITCH,
        )
        self.bayesian_switch.deselect()
        self.bayesian_switch.grid(
            row=3,
            column=1,
        )

    # ============ FORWARD ============

    def setup_forward(self):

        self.forward_label = customtkinter.CTkLabel(
            text="Forward only:",
            **self.RIGHT_LABEL_KW,
        )
        self.forward_label.grid(
            row=4,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.forward_bool = customtkinter.BooleanVar()
        self.forward_button_title = customtkinter.StringVar(
            value="False",
        )
        self.forward_switch = customtkinter.CTkSwitch(
            master=self.frame_right,
            textvariable=self.forward_button_title,
            command=self.forward_callback,
            onvalue=True,
            offvalue=False,
            **KW_SWITCH,
        )
        self.forward_switch.deselect()
        self.forward_switch.grid(
            row=4,
            column=1,
        )

    # ============ PARALLEL SAMPLES ============

    def setup_parallel_samples(self):

        self.parallel_label = customtkinter.CTkLabel(
            text="Samples in parallel:",
            **self.RIGHT_LABEL_KW,
        )
        self.parallel_label.grid(
            row=5,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.frame_parallel_samples = customtkinter.CTkFrame(
            master=self.frame_right,
            fg_color="#2A2D2E",
        )
        self.frame_parallel_samples.grid(
            row=5,
            column=1,
            sticky="e",
        )

        PARALLEL_SAMPLES_MIN = 1
        PARALLEL_SAMPLES_DEFAULT = 1
        PARALLEL_SAMPLES_MAX = 20

        self.parallel_samples_slider = customtkinter.CTkSlider(
            master=self.frame_parallel_samples,
            command=self.parallel_samples_slider_callback,
            from_=PARALLEL_SAMPLES_MIN,
            to=PARALLEL_SAMPLES_MAX,
            number_of_steps=PARALLEL_SAMPLES_MAX - PARALLEL_SAMPLES_MIN,
            width=100,
            **KW_SLIDER,
        )
        self.parallel_samples_slider.set(PARALLEL_SAMPLES_DEFAULT)
        self.parallel_samples_slider.grid(
            row=0,
            column=0,
            sticky="e",
        )

        self.parallel_samples_value = customtkinter.CTkLabel(
            master=self.frame_parallel_samples,
            width=20,
        )
        self.parallel_samples_value.grid(
            row=0,
            column=1,
            pady=12,
            padx=0,
        )

        self.parallel_samples_slider_callback(PARALLEL_SAMPLES_DEFAULT)

    # ============ CORES PR. SAMPLE ============

    def setup_cores_per_sample(self):

        self.cores_per_sample_label = customtkinter.CTkLabel(
            text="Cores pr. sample:",
            **self.RIGHT_LABEL_KW,
        )
        self.cores_per_sample_label.grid(
            row=6,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.frame_cores_per_sample = customtkinter.CTkFrame(
            master=self.frame_right,
            fg_color="#2A2D2E",
        )
        self.frame_cores_per_sample.grid(
            row=6,
            column=1,
            sticky="e",
        )

        PARALLEL_CORES_PER_SAMPLE_MIN = 1
        PARALLEL_CORES_PER_SAMPLE_DEFAULT = 1
        PARALLEL_CORES_PER_SAMPLE_MAX = 20

        self.cores_per_sample_slider = customtkinter.CTkSlider(
            master=self.frame_cores_per_sample,
            command=self.cores_per_sample_slider_callback,
            from_=PARALLEL_CORES_PER_SAMPLE_MIN,
            to=PARALLEL_CORES_PER_SAMPLE_MAX,
            number_of_steps=PARALLEL_CORES_PER_SAMPLE_MAX
            - PARALLEL_CORES_PER_SAMPLE_MIN,
            width=100,
            **KW_SLIDER,
        )
        self.cores_per_sample_slider.set(PARALLEL_CORES_PER_SAMPLE_DEFAULT)
        self.cores_per_sample_slider.grid(
            row=0,
            column=0,
            sticky="e",
        )

        self.cores_per_sample_value = customtkinter.CTkLabel(
            master=self.frame_cores_per_sample,
            width=20,
        )
        self.cores_per_sample_value.grid(
            row=0,
            column=1,
            pady=12,
            padx=0,
        )

        self.cores_per_sample_slider_callback(PARALLEL_CORES_PER_SAMPLE_DEFAULT)

    # ============ PREFIX ============

    def setup_prefix(self):

        self.prefix_label = customtkinter.CTkLabel(
            text="Prefix:",
            **self.RIGHT_LABEL_KW,
        )
        self.prefix_label.grid(
            row=7,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.prefix_entry = customtkinter.CTkEntry(
            master=self.frame_right,
            placeholder_text="",
            **KW_ENTRY_GOOD_COLORS,
        )
        self.prefix_entry.grid(
            row=7,
            column=1,
        )

    # ============ SUFFIX ============

    def setup_suffix(self):

        self.suffix_label = customtkinter.CTkLabel(
            text="Suffix:",
            **self.RIGHT_LABEL_KW,
        )
        self.suffix_label.grid(
            row=8,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.suffix_entry = customtkinter.CTkEntry(
            master=self.frame_right,
            placeholder_text="",
            **KW_ENTRY_GOOD_COLORS,
        )
        self.suffix_entry.grid(
            row=8,
            column=1,
        )

    # ============ LONG NAME ============

    def setup_long_name(self):

        self.long_name_label = customtkinter.CTkLabel(
            text="Long name:",
            **self.RIGHT_LABEL_KW,
        )
        self.long_name_label.grid(
            row=9,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.long_name_bool = customtkinter.BooleanVar()
        self.long_name_button_title = customtkinter.StringVar(
            value="False",
        )
        self.long_name_switch = customtkinter.CTkSwitch(
            master=self.frame_right,
            textvariable=self.long_name_button_title,
            command=self.long_name_callback,
            onvalue=True,
            offvalue=False,
            **KW_SWITCH,
        )
        self.long_name_switch.deselect()
        self.long_name_switch.grid(
            row=9,
            column=1,
        )

    # ============ CONFIG NAME ============

    def setup_config_name(self):

        self.config_name_label = customtkinter.CTkLabel(
            text="Config Name:",
            **self.RIGHT_LABEL_KW,
        )
        self.config_name_label.grid(
            row=10,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.config_button = customtkinter.CTkButton(
            master=self.frame_right,
            text="Select file",
            command=self.config_callback,
            **KW_BUTTON_BASE,
            **KW_BUTTON_BAD_COLORS,
        )
        self.config_button.grid(
            row=10,
            column=1,
            pady=12,
            padx=10,
        )

        self.config_is_okay = False
        self.config_file_path = None

    # ============ OUTPUT DIRECTORY ============

    def setup_output_dir(self):

        self.output_dir_label = customtkinter.CTkLabel(
            text="Output Directory",
            **self.RIGHT_LABEL_KW,
        )
        self.output_dir_label.grid(
            row=12,
            column=0,
            **RIGHT_LABEL_GRID_KW,
        )

        self.output_dir_button = customtkinter.CTkButton(
            master=self.frame_right,
            text=format_directory(cli_utils.output_dir_default),
            command=self.output_dir_callback,
            **KW_BUTTON_BASE,
            **KW_BUTTON_GOOD_COLORS,
        )
        self.output_dir_button.grid(
            row=12,
            column=1,
        )

        self.output_dir_path = cli_utils.output_dir_default

    # ============ SAVE CONFIG ============

    def setup_save_config(self):

        self.frame_save_config = customtkinter.CTkFrame(
            master=self,
            fg_color="#212325",
        )
        self.frame_save_config.grid(
            row=3,
            column=0,
        )

        self.save_config_button = customtkinter.CTkButton(
            master=self.frame_save_config,
            text="Save Config",
            command=self.save_config_callback,
        )
        self.save_config_button.grid(
            row=0,
            column=0,
            pady=12,
            padx=10,
        )

        # ============ BAM FILE (FILE) ============

    # def init_bam(self):

    def bam_callback(self):
        filepaths = filedialog.askopenfilenames()
        # filepath = filedialog.askopenfilename()

        if filepaths == "":
            return

        elif len(filepaths) == 1:
            filepath = filepaths[0]
            text = path_to_text(filepath)
            self.bam_button.configure(text=text)
            self.bam_file_path = [Path(filepath)]

        else:
            # self.bam_file_string.set(str(filepaths))
            self.bam_button.configure(text="File paths set")
            self.bam_file_path = [Path(path) for path in filepaths]

        self.bam_is_okay = True
        self.bam_button.configure(**KW_BUTTON_GOOD_COLORS)

    # ============ DAMAGE MODE (ENUM) ============

    def damage_mode_collback(self, choice):

        if choice == cli_utils.DAMAGE_MODE.LCA:
            self.damage_mode_value = cli_utils.DAMAGE_MODE.LCA
        elif choice == cli_utils.DAMAGE_MODE.LOCAL:
            self.damage_mode_value = cli_utils.DAMAGE_MODE.LOCAL
        elif choice == cli_utils.DAMAGE_MODE.GLOBAL:
            self.damage_mode_value = cli_utils.DAMAGE_MODE.GLOBAL

        if choice == cli_utils.DAMAGE_MODE.LCA:
            lca_state = "normal"
            text_color = "gray90"
            slider = KW_SLIDER
            lca_rank_menu_kw = KW_OPTIONS_MENU
            switch = KW_SWITCH
            self.is_lca = True

        else:
            lca_state = "disabled"
            text_color = "gray40"
            slider = KW_SLIDER_DISABLED
            lca_rank_menu_kw = KW_OPTIONS_MENU_DISABLED
            switch = KW_SWITCH_DISABLED
            self.is_lca = False

        texts = [
            self.names_label,
            self.nodes_label,
            self.acc2tax_label,
            self.min_mapping_quality_label,
            self.min_mapping_quality_value,
            self.custom_database_label,
            self.lca_rank_label,
        ]

        for text in texts:
            text.configure(text_color=text_color)

        states = [
            self.names_button,
            self.nodes_button,
            self.acc2tax_button,
            self.min_mapping_quality_slider,
            self.custom_database_switch,
            self.lca_rank_menu,
        ]

        for state in states:
            state.configure(state=lca_state)

        self.min_mapping_quality_slider.configure(**slider)
        self.lca_rank_menu.configure(**lca_rank_menu_kw)
        self.custom_database_switch.configure(**switch)

        if choice == cli_utils.DAMAGE_MODE.LCA:
            self.names_button.configure(**self.names_button_colors)
            self.nodes_button.configure(**self.nodes_button_colors)
            self.acc2tax_button.configure(**self.acc2tax_button_colors)
        else:
            self.names_button.configure(**KW_BUTTON_GOOD_COLORS)
            self.nodes_button.configure(**KW_BUTTON_GOOD_COLORS)
            self.acc2tax_button.configure(**KW_BUTTON_GOOD_COLORS)

        self.check_similarity_scores()

    # ============ NAMES FILE (FILE) ============

    def names_callback(self):
        filepath = filedialog.askopenfilename()
        if filepath != "":
            self.names_file_path = Path(filepath)
            text = path_to_text(filepath)
            self.names_button.configure(text=text)
            self.names_button.configure(**KW_BUTTON_GOOD_COLORS)
            self.names_is_okay = True
            self.names_button_colors = KW_BUTTON_GOOD_COLORS

    # ============ NODES FILE (FILE) ============

    def nodes_callback(self):
        filepath = filedialog.askopenfilename()
        if filepath != "":
            self.nodes_file_path = Path(filepath)
            text = path_to_text(filepath)
            self.nodes_button.configure(text=text)
            self.nodes_button.configure(**KW_BUTTON_GOOD_COLORS)
            self.nodes_is_okay = True
            self.nodes_button_colors = KW_BUTTON_GOOD_COLORS

    # ============ ACC2TAX FILE (FILE) ============

    def acc2tax_callback(self):
        filepath = filedialog.askopenfilename()
        if filepath != "":
            self.acc2tax_file_path = Path(filepath)
            text = path_to_text(filepath)
            self.acc2tax_button.configure(text=text)
            self.acc2tax_button.configure(**KW_BUTTON_GOOD_COLORS)
            self.acc2tax_is_okay = True
            self.acc2tax_button_colors = KW_BUTTON_GOOD_COLORS

    #############

    def config_callback(self):
        filepath = filedialog.asksaveasfilename()
        if filepath != "":

            if not ("yaml" in filepath or "yml" in filepath):
                filepath += ".yaml"

            self.config_file_path = Path(filepath)
            text = path_to_text(filepath)
            self.config_button.configure(text=text)
            self.config_button.configure(**KW_BUTTON_GOOD_COLORS)
            self.config_is_okay = True
            self.config_button_colors = KW_BUTTON_GOOD_COLORS

    # ============ MAPPING QUALITY (INTEGER) ============

    def min_mapping_quality_slider_callback(self, value):
        self.min_mapping_quality_value.configure(text=f"= {int(value):>2d}")

    # ============ CUSTOM DATABASE (BOOL) ============

    def custom_database_callback(self):
        state = self.custom_database_switch.get()
        self.custom_database_button_title.set(str(state))
        self.custom_database_bool.set(state)

    # ============ LCA RANK (ENUM) ============

    def lca_rank_callback(self, choice):
        self.lca_rank_string.set(choice if choice != "none" else cli_utils.RANKS.none)

    # ============ MAX POSITION (INTEGER) ============

    def max_position_slider_callback(self, value):
        self.max_position_value.configure(text=f"= {int(value):>2d}")

    # ============ BAYESIAN (BOOL) ============

    def bayesian_callback(self):
        state = self.bayesian_switch.get()
        self.bayesian_button_title.set(str(state))
        self.bayesian_bool.set(state)

    # ============ FORWARD (BOOL) ============

    def forward_callback(self):
        state = self.forward_switch.get()
        self.forward_button_title.set(str(state))
        self.forward_bool.set(state)

    # ============ PARALLEL CORES (SLIDER) ============

    def parallel_samples_slider_callback(self, value):
        self.parallel_samples_value.configure(text=f"= {int(value):>2d}")

    # ============ CORES PER SAMPLE (SLIDER) ============

    def cores_per_sample_slider_callback(self, value):
        self.cores_per_sample_value.configure(text=f"= {int(value):>2d}")

    # ============ LONG NAME (ENTRY) ============

    def long_name_callback(self):
        state = self.long_name_switch.get()
        self.long_name_button_title.set(str(state))
        self.long_name_bool.set(state)

    # ============ CHECK SIMILARITY SCORE ============

    def check_similarity_scores(self):

        if self.is_lca:
            text_color = "gray90"
            self.similarity_score_label.configure(text_color=text_color)
            self.similarity_score_to.configure(text_color=text_color)
            self.similarity_score_min.configure(state="normal")
            self.similarity_score_max.configure(state="normal")

            try:
                min_score = float(self.similarity_score_min.get())
                max_score = float(self.similarity_score_max.get())
                is_float = True
            except ValueError:
                is_float = False

            if is_float and (0 <= min_score <= max_score <= 1):
                self.similarity_score_min.configure(**KW_ENTRY_GOOD_COLORS)
                self.similarity_score_max.configure(**KW_ENTRY_GOOD_COLORS)
                self.similarity_is_ok = True

            else:
                self.similarity_score_min.configure(**KW_ENTRY_BAD_COLORS)
                self.similarity_score_max.configure(**KW_ENTRY_BAD_COLORS)
                self.similarity_is_ok = False
        else:
            text_color = "gray40"
            self.similarity_score_label.configure(text_color=text_color)
            self.similarity_score_to.configure(text_color=text_color)
            self.similarity_score_min.configure(state="disabled", **KW_ENTRY_DISABLED)
            self.similarity_score_max.configure(state="disabled", **KW_ENTRY_DISABLED)

    def check_similarity_scores_background(self, interval_sec):
        # run forever
        while True:
            # block for the interval
            sleep(interval_sec)
            # perform the task
            self.check_similarity_scores()

    def start_similarity_scores_background_thread(self):
        # create and start the daemon thread
        daemon = Thread(
            target=self.check_similarity_scores_background,
            args=(0.05,),
            daemon=True,
            name="Similarity-score-checking",
        )
        daemon.start()

    # ============ CHECK MIN READS  ============

    def check_minimum_reads(self):

        try:
            min_reads = int(self.min_reads_value.get())
            is_int = True
        except ValueError:
            is_int = False

        if is_int and (0 <= min_reads):
            self.min_reads_is_ok = True
            self.min_reads_entry.configure(**KW_ENTRY_GOOD_COLORS)
        else:
            self.min_reads_entry.configure(**KW_ENTRY_BAD_COLORS)
            self.min_reads_is_ok = False

    def check_minimum_reads_background(self, interval_sec):
        # run forever
        while True:
            # block for the interval
            sleep(interval_sec)
            # perform the task
            self.check_minimum_reads()

    def start_minimum_reads_background_thread(self):
        # create and start the daemon thread
        daemon = Thread(
            target=self.check_minimum_reads_background,
            args=(0.05,),
            daemon=True,
            name="minimum-reads-checking",
        )
        daemon.start()

    # ============ OUTPUT DIR (DIRECTORY) ============

    def output_dir_callback(self):
        # get a directory path by user
        filepath = filedialog.askdirectory()
        self.output_dir_path = Path(filepath)

        text = format_directory(Path(filepath))
        self.output_dir_button.configure(text=text)

    def create_popup_window(self, text):

        window_popup = customtkinter.CTkToplevel(self)
        window_popup.title("metaDMG - Error")

        width = 365
        height = 110
        x, y = self.get_center_coordinates(width, height)

        window_popup.geometry(f"{width}x{height}+{x}+{y}")

        label = customtkinter.CTkLabel(
            window_popup,
            text=f"Error occurred in config creation. \nPlease {text}.",
        )

        label.grid(
            row=0,
            column=0,
            padx=80,
            pady=12,
        )

        def close_popup_callback():
            window_popup.destroy()

        button = customtkinter.CTkButton(
            master=window_popup,
            text="Close",
            command=close_popup_callback,
        )
        button.grid(
            row=1,
            column=0,
            padx=80,
            pady=12,
        )

    # ============ PRINT ============

    def config_is_good(self):

        self.check_similarity_scores()
        self.check_minimum_reads()

        if not self.bam_is_okay:
            self.create_popup_window("fix BAM file")
            return False

        if self.is_lca and (not self.names_is_okay):
            self.create_popup_window("fix names file")
            return False

        if self.is_lca and (not self.nodes_is_okay):
            self.create_popup_window("fix nodes file")
            return False

        if self.is_lca and (not self.acc2tax_is_okay):
            self.create_popup_window("fix acc2tax file")
            return False

        if self.is_lca and (not self.similarity_is_ok):
            self.create_popup_window("fix similarity score")
            return False

        if not self.min_reads_is_ok:
            self.create_popup_window("fix minimum number of reads")
            return False

        if not self.config_is_okay:
            self.create_popup_window("fix config file name")
            return False

        return True

    def get_config(self):

        if not self.config_is_good():
            return None

        from metaDMG import __version__

        config = cli_utils.get_config_dict(
            samples=self.bam_file_path,
            damage_mode=self.damage_mode_value,
            #
            names=self.names_file_path,
            nodes=self.nodes_file_path,
            acc2tax=self.acc2tax_file_path,
            min_similarity_score=float(self.similarity_score_min.get()),
            max_similarity_score=float(self.similarity_score_max.get()),
            min_edit_dist=None,
            max_edit_dist=None,
            min_mapping_quality=int(self.min_mapping_quality_slider.get()),
            custom_database=self.custom_database_bool.get(),
            lca_rank=self.lca_rank_string.get(),
            #
            metaDMG_cpp="./metaDMG-cpp",
            max_position=int(self.max_position_slider.get()),
            min_reads=int(self.min_reads_value.get()),
            weight_type=1,
            forward_only=self.forward_bool.get(),
            bayesian=self.bayesian_bool.get(),
            output_dir=self.output_dir_path,
            parallel_samples=int(self.parallel_samples_slider.get()),
            cores_per_sample=int(self.cores_per_sample_slider.get()),
            config_file=self.config_file_path,
            sample_prefix=self.prefix_entry.get(),
            sample_suffix=self.suffix_entry.get(),
            long_name=self.long_name_bool.get(),
            __version__=__version__,
        )

        return config

    def save_config_file(self, config, overwrite_config=False):
        if self.verbose:
            print("")
            print("Config file to save:")
            print(config)
        cli_utils.save_config_file(
            config=config,
            config_file=self.config_file_path,
            overwrite_config=overwrite_config,
            verbose=self.verbose,
        )

    def make_overwrite_window(self, config):

        window_overwrite = customtkinter.CTkToplevel(self)
        window_overwrite.title("metaDMG - Computing")

        width = 365
        height = 110
        x, y = self.get_center_coordinates(width, height)

        window_overwrite.geometry(f"{width}x{height}+{x}+{y}")

        label = customtkinter.CTkLabel(
            window_overwrite,
            text="Config file already exists. Do you want to overwrite it?",
        )

        label.grid(
            row=0,
            column=0,
            columnspan=2,
            padx=20,
            pady=12,
        )

        def no_overwrite_callback():
            window_overwrite.destroy()

        def overwrite_callback():
            self.save_config_file(config=config, overwrite_config=True)
            window_overwrite.destroy()

        button_yes = customtkinter.CTkButton(
            master=window_overwrite,
            text="Yes",
            command=overwrite_callback,
        )
        button_yes.grid(
            row=1,
            column=0,
            padx=20,
            pady=12,
        )

        button_no = customtkinter.CTkButton(
            master=window_overwrite,
            text="No",
            command=no_overwrite_callback,
        )
        button_no.grid(
            row=1,
            column=1,
            padx=20,
            pady=12,
        )

    def save_config_callback(self):

        config = self.get_config()
        if config is None:
            return

        if self.config_file_path.exists():
            self.make_overwrite_window(config)
        else:
            self.save_config_file(config)

    # ============ OTHER ============

    def on_closing(self, event=0):
        self.destroy()

    def get_center_coordinates(self, width, height):
        # get screen width and height
        ws = self.winfo_screenwidth()  # width of the screen
        hs = self.winfo_screenheight()  # height of the screen

        # calculate x and y coordinates for the Tk root window
        x = (ws / 2) - (width / 2)
        y = (hs / 2) - (height / 2)

        return int(x), int(y)


if __name__ == "__main__":
    gui = Gui(verbose=True)
    gui.mainloop()



#%%
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.stats import betabinom as sp_betabinom

from metaDMG.viz import viz_utils


def sort_dataframe(df):
    samples_ordered = list(
        df.groupby("sample")
        .sum("N_reads")
        .sort_values("N_reads", ascending=False)
        .index
    )

    # Create the dictionary that defines the order for sorting
    sorterIndex = dict(zip(samples_ordered, range(len(samples_ordered))))

    # Generate a rank column that will be used to sort
    # the dataframe numerically
    df["sample_rank"] = df["sample"].map(sorterIndex)

    df = df.sort_values(
        ["sample_rank", "tax_id"],
        ascending=[True, True],
        inplace=False,
    ).drop(columns="sample_rank")

    return df


def clip_df(df, column):
    if column in df.columns:
        df["_" + column] = df[column]  # save original data _column
        df[column] = np.clip(df[column], a_min=0, a_max=None)


def pd_wide_to_long_forward_reverse(group_wide, sep, direction):
    stub_names = ["k", "N", "f"]
    group_long = pd.wide_to_long(
        group_wide,
        stubnames=stub_names,
        i="tax_id",
        j="|x|",
        sep=sep,
    )[stub_names]
    group_long["direction"] = direction
    return group_long.reset_index()


def wide_to_long_df(group_wide):

    group_long_forward = pd_wide_to_long_forward_reverse(
        group_wide,
        sep="+",
        direction="Forward",
    )

    try:
        group_long_reverse = pd_wide_to_long_forward_reverse(
            group_wide,
            sep="-",
            direction="Reverse",
        )

        group_long = pd.concat([group_long_forward, group_long_reverse])
        # group_long.loc[:, ["k", "N"]] = group_long.loc[:, ["k", "N"]].astype(int)

    # happens when forward only
    except ValueError:
        group_long = group_long_forward

    group_long["sample"] = group_wide["sample"].iloc[0]

    return group_long


# def remove_LCA_columns(columns):
#     remove_cols = ["tax_name", "tax_rank", "N_alignments"]
#     return [col for col in columns if col not in remove_cols]


def correct_for_non_LCA(df):
    if "N_alignments" in df.columns:
        df.loc[:, "N_alignments"] = df["N_alignments"].fillna(np.nan)
    else:
        df.loc[:, "N_alignments"] = np.nan

    if "tax_name" in df.columns:
        df.loc[:, "tax_name"] = df["tax_name"].fillna("NO TAX NAME")
    else:
        df.loc[:, "tax_name"] = "NO TAX NAME"

    if "tax_rank" in df.columns:
        df.loc[:, "tax_rank"] = df["tax_rank"].fillna("NO TAX RANK")
    else:
        df.loc[:, "tax_rank"] = "NO TAX RANK"

    if "tax_path" in df.columns:
        df.loc[:, "tax_path"] = df["tax_path"].fillna("NO TAX PATH")
    else:
        df.loc[:, "tax_path"] = "NO TAX PATH"

    return df


#%%


def compute_variance_scaling(df, phi_string):
    phi = df[phi_string]
    if "N_x=1_reverse" in df.columns:
        N_x = [df["N_x=1_forward"], df["N_x=1_reverse"]]
    else:
        N_x = [df["N_x=1_forward"]]
    N = np.mean(N_x)
    return (phi + N) / (phi + 1)


#%%


class VizResults:
    def __init__(self, results_dir):
        self.results_dir = Path(results_dir)
        self._load_df_results()
        self._set_cmap()
        self._set_hover_info()

    def _load_parquet_file(self, results_dir):
        df = pd.read_parquet(results_dir)
        return correct_for_non_LCA(df)

    def _load_df_results(self):
        df = self._load_parquet_file(self.results_dir)
        df = sort_dataframe(df)

        # force tax_id to be categorical strings.
        # XXX remove in final version
        df.loc[:, "tax_id"] = df["tax_id"].astype("str").astype("category")

        for column in ["lambda_LR", "forward_lambda_LR", "reverse_lambda_LR"]:
            clip_df(df, column)

        Bayesian = any(["Bayesian" in column for column in df.columns]) and (
            not any(df["Bayesian_z"].isna())
        )
        self.Bayesian = Bayesian

        df["D_max_significance"] = df["D_max"] / df["D_max_std"]
        df["rho_Ac_abs"] = np.abs(df["rho_Ac"])
        df["variance_scaling"] = compute_variance_scaling(df, phi_string="phi")

        if Bayesian:
            df["Bayesian_D_max_significance"] = (
                df["Bayesian_D_max"] / df["Bayesian_D_max_std"]
            )
            df["Bayesian_rho_Ac_abs"] = np.abs(df["Bayesian_rho_Ac"])
            df["Bayesian_variance_scaling"] = compute_variance_scaling(
                df, phi_string="Bayesian_phi"
            )

        log_columns = [
            "N_reads",
            "N_alignments",
            "lambda_LR",
            "phi",
            "k_sum_total",
            "N_sum_total",
        ]

        for column in log_columns:
            log_column = "log_" + column
            df.loc[:, log_column] = np.log10(1 + df[column])

        if np.isnan(df["asymmetry"]).all() and not "forward_A" in df.columns:
            self.forward_only = True
        else:
            self.forward_only = False

        self.df = df

        self.all_tax_ids = set(self.df["tax_id"].unique())
        self.all_tax_names = set(self.df["tax_name"].unique())  # if with_LCA else set()
        self.all_tax_ranks = set(self.df["tax_rank"].unique())  # if with_LCA else set()
        self.samples = list(self.df["sample"].unique())
        self.columns = list(self.df.columns)
        self.set_marker_size(variable="N_reads", function="sqrt", slider=30)

    def set_marker_size(self, variable="N_reads", function="sqrt", slider=30):

        d_functions = {
            "constant": np.ones_like,
            "linear": lambda x: x,
            "sqrt": np.sqrt,
            "log10": np.log10,
        }

        self.df.loc[:, "size"] = d_functions[function](self.df[variable])

        self.max_of_size = np.max(self.df["size"])
        self.marker_size = slider

    def filter(self, filters, *, rank=None):

        query = ""
        for column, filter in filters.items():

            if filter is None:
                continue

            elif column == "samples":
                query += f"(sample in {filter}) & "

            elif column == "sample":
                query += f"(sample == '{filter}') & "

            elif column == "tax_id":
                query += f"(tax_id == '{filter}') & "

            elif column == "tax_ids":
                query += f"(tax_id in {filter}) & "

            elif column == "tax_rank":
                query += f"(tax_rank == {filter}) & "

            elif column == "tax_ranks":
                query += f"(tax_rank in {filter}) & "

            elif column == "tax_name":
                query += f"(tax_name == {filter}) & "

            elif column == "tax_names":
                query += f"(tax_name in {filter}) & "

            elif column == "tax_path":
                rank = filter

            else:
                low, high = filter
                if viz_utils.is_log_transform_column(column):
                    low = viz_utils.log_transform_slider(low)
                    high = viz_utils.log_transform_slider(high)
                query += f"({low} <= {column} <= {high}) & "

        query = query[:-2]
        # print(query)

        df_out = self.df.query(query)

        if rank is not None:
            mask = df_out["tax_path"].str.lower().str.contains(rank.lower())
            df_out = df_out.loc[mask]

        return df_out

    def filter_tax_path(self, rank):
        mask = self.df["tax_path"].str.contains(rank)
        return self.df.loc[mask]

    def _set_cmap(self):
        # https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express
        # blue, orange, green, red, purple, brown, pink, grey, camouflage, turquoise
        # cmap = px.colors.qualitative.D3
        # cmap taken from http://www.cookbook-r.com/Graphs/Colors_%28ggplot2%29/
        cmap = [
            "#0072B2",
            "#D55E00",
            "#009E73",
            "#CC79A7",
            "#E69F00",
            "#56B4E9",
            "#F0E442",
        ]
        N_cmap = len(cmap)

        groupby = self.df.groupby("sample", sort=False)

        symbol_counter = 0
        d_cmap = {}
        d_symbols = {}
        markers = ["o", "s", "^", "v", "<", ">", "d"]
        d_markers = {}
        for i, (name, _) in enumerate(groupby):

            if (i % N_cmap) == 0 and i != 0:
                symbol_counter += 1

            d_cmap[name] = cmap[i % N_cmap]
            d_symbols[name] = symbol_counter % 44  # max Plotly symbol number
            d_markers[name] = markers[symbol_counter % len(markers)]

        self.cmap = cmap
        self.d_cmap = d_cmap
        self.d_symbols = d_symbols
        self.d_markers = d_markers

        self.d_cmap_fit = {"Forward": cmap[0], "Reverse": cmap[3], "Fit": cmap[2]}

    def _set_hover_info(self):

        placeholder = "_XXX_"

        self.custom_data_columns = [
            "sample",
            "tax_name",
            "tax_rank",
            "tax_id",
            # Bayesian Fits
            # Frequentist fits
            "lambda_LR",
            "D_max",
            "D_max_std",
            "q",
            "q_std",
            "phi",
            "phi_std",
            "asymmetry",
            "rho_Ac",
            # Counts
            "N_reads",
            "N_alignments",
            "N_sum_total",
            "k_sum_total",
        ]

        custom_data_columns_Bayesian = [
            "Bayesian_z",
            "Bayesian_D_max",
            "Bayesian_D_max_std",
            "Bayesian_q",
            "Bayesian_q_std",
            "Bayesian_phi",
            "Bayesian_phi_std",
            "Bayesian_rho_Ac",
        ]

        self.hovertemplate = (
            "<b>%{customdata[_XXX_]}</b><br><br>"
            "<b>Tax</b>: <br>"
            "    Name: %{customdata[_XXX_]} <br>"
            "    Rank: %{customdata[_XXX_]} <br>"
            "    ID:   %{customdata[_XXX_]} <br><br>"
            "<b>MAP results</b>: <br>"
            "    LR:       %{customdata[_XXX_]:9.2f} <br>"
            "    D max:    %{customdata[_XXX_]:9.2f} ± %{customdata[_XXX_]:.2f} <br>"
            "    q:        %{customdata[_XXX_]:9.2f} ± %{customdata[_XXX_]:.2f} <br>"
            "    phi:      %{customdata[_XXX_]:.3s} ± %{customdata[_XXX_]:.3s} <br>"
            "    asymmetry:%{customdata[_XXX_]:9.3f} <br>"
            "    rho_Ac:   %{customdata[_XXX_]:9.3f} <br><br>"
            "<b>Counts</b>: <br>"
            "    N reads:     %{customdata[_XXX_]:.3s} <br>"
            "    N alignments:%{customdata[_XXX_]:.3s} <br>"
            "    N sum total: %{customdata[_XXX_]:.3s} <br>"
            "    k sum total: %{customdata[_XXX_]:.3s} <br>"
            "<extra></extra>"
        )

        hovertemplate_Bayesian = (
            "<b>Fit results</b>: <br>"
            "    z:        %{customdata[_XXX_]:9.2f} <br>"
            "    D max:    %{customdata[_XXX_]:9.2f} ± %{customdata[_XXX_]:.2f} <br>"
            "    q:        %{customdata[_XXX_]:9.2f} ± %{customdata[_XXX_]:.2f} <br>"
            "    phi:      %{customdata[_XXX_]:.3s} ± %{customdata[_XXX_]:.3s} <br>"
            "    rho_Ac:   %{customdata[_XXX_]:9.3f} <br><br>"
        )

        if self.forward_only:
            index = self.hovertemplate.find("<b>MAP results</b>: <br>")
            self.hovertemplate = (
                self.hovertemplate[:index]
                + "<b>Forward only! </b><br><br>"
                + self.hovertemplate[index:]
            )

        # if Bayesian fits, include these results
        if self.Bayesian:
            index = self.custom_data_columns.index("lambda_LR")
            self.custom_data_columns[index:index] = custom_data_columns_Bayesian

            index = self.hovertemplate.find("<b>MAP results</b>: <br>")
            self.hovertemplate = (
                self.hovertemplate[:index]
                + hovertemplate_Bayesian
                + self.hovertemplate[index:]
            )

        # fill in the templates with data
        data_counter = 0
        i = 0
        while True:
            if self.hovertemplate[i : i + len(placeholder)] == placeholder:
                # break
                s_new = self.hovertemplate[:i]
                s_new += str(data_counter)
                s_new += self.hovertemplate[i + len(placeholder) :]
                self.hovertemplate = s_new
                data_counter += 1
            i += 1
            if i >= len(self.hovertemplate):
                break

        # if not self.with_LCA:
        #     self.custom_data_columns = remove_LCA_columns(self.custom_data_columns)
        # self.customdata = self.df[self.custom_data_columns]
        self.hovertemplate_fit = (
            "Fit: <br>D(x) = %{y:.3f} ± %{error_y.array:.3f}<br>" "<extra></extra>"
        )

    def parse_click_data(self, click_data, column):
        try:
            index = self.custom_data_columns.index(column)
            value = click_data["points"][0]["customdata"][index]
            return value

        except Exception as e:
            raise e

    def get_single_count_group(self, sample, tax_id, forward_reverse=""):
        query = f"sample == '{sample}' & tax_id == '{tax_id}'"
        group_wide = self.df.query(query)
        group = wide_to_long_df(group_wide)

        if forward_reverse.lower() == "forward":
            return group.query(f"direction=='Forward'")
        elif forward_reverse.lower() == "reverse":
            return group.query(f"direction=='Reverse'")
        else:
            return group

    def get_single_fit_prediction(self, sample, tax_id, forward_reverse=""):
        query = f"sample == '{sample}' & tax_id == '{tax_id}'"
        ds = self.df.query(query)
        if len(ds) != 1:
            raise AssertionError(f"Something wrong here, got: {ds}")

        if self.forward_only:
            if forward_reverse.lower() == "reverse":
                return "FORWARD ONLY"
            else:
                forward_reverse = ""

        group = self.get_single_count_group(sample, tax_id, forward_reverse)

        if forward_reverse.lower() == "forward":
            prefix = "forward_"
        elif forward_reverse.lower() == "reverse":
            prefix = "reverse_"
        else:
            if self.Bayesian:
                prefix = "Bayesian_"
            else:
                prefix = ""

        A = getattr(ds, f"{prefix}A").values
        q = getattr(ds, f"{prefix}q").values
        c = getattr(ds, f"{prefix}c").values
        phi = getattr(ds, f"{prefix}phi").values

        max_position = viz_utils.get_max_position_from_group(group)

        abs_x = group["|x|"].values[:max_position]
        N = group["N"].values[:max_position]

        Dx = A * (1 - q) ** (abs_x - 1) + c

        alpha = Dx * phi
        beta = (1 - Dx) * phi

        dist = sp_betabinom(n=N, a=alpha, b=beta)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            std = dist.std() / N

        std[np.isnan(std)] = 0

        text = self._ds_to_fit_text(ds)

        d_out = {"mu": Dx, "std": std, "Dx": Dx, "|x|": abs_x, "text": text}

        return d_out

    def _ds_to_fit_text(self, ds):

        d_columns_latex = viz_utils.get_d_columns_latex(self)[0]

        sanitize = (
            lambda s: d_columns_latex[s]
            .strip(r"$")
            .replace(r"\text{(MAP)}", r"")
            .replace(r"\text", r"\mathrm")
        )

        text = r"$\mathrm{Bayesian}" if self.Bayesian else r"$\mathrm{MAP}"
        text += r"\,\, \mathrm{fit}$" + "\n\n"

        D_max_col = "Bayesian_D_max" if self.Bayesian else "D_max"
        D_max_str = sanitize(D_max_col)
        D_max = ds[D_max_col].iloc[0]
        D_max_std = ds[D_max_col + "_std"].iloc[0]
        text += (
            "$" + D_max_str + f" = {D_max:.3f} " + r"\pm" + f" {D_max_std:.3f}" + r"$"
        )
        text += "\n"

        fitqual_col = "Bayesian_z" if self.Bayesian else "lambda_LR"
        fitqual_str = sanitize(fitqual_col)
        fitqual = ds[fitqual_col].iloc[0]
        text += "$" + fitqual_str + f" = {fitqual:.2f} " + r"$"
        text += "\n"

        phi_col = "Bayesian_phi" if self.Bayesian else "phi"
        phi_str = sanitize(phi_col)
        phi = viz_utils.human_format(ds[phi_col].iloc[0])
        phi_std = viz_utils.human_format(ds[phi_col + "_std"].iloc[0])
        text += "$" + phi_str + f" = {phi} " + r"\pm" + f" {phi_std}" + r"$"
        text += "\n"

        q_col = "Bayesian_q" if self.Bayesian else "q"
        q_str = sanitize(q_col)
        q = ds[q_col].iloc[0]
        q_std = ds[q_col + "_std"].iloc[0]
        text += "$" + q_str + f" = {q:.2f} " + r"\pm" + f" {q_std:.2f}" + r"$"
        # text += "\n"

        return text



from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt

from metaDMG.viz import app, viz_utils


def start_dashboard(
    results_dir: Optional[Path] = None,
    debug: bool = False,
    server: bool = False,
    host: str = "0.0.0.0",
    port: int = 8050,
):

    if results_dir is None:
        raise Exception(f"Has to be specified.")

    plt.switch_backend("Agg")

    if not (debug or server):
        viz_utils.open_browser_in_background(port)

    dashboard_app = app.get_app(results_dir)

    dashboard_app.run_server(
        debug=debug,
        host=host,
        port=str(port),
        use_reloader=False,
    )
